import unittest
from time import time_ns
from unittest.mock import patch
from uuid import uuid1

from uuid6 import DraftUUID, uuid6, uuid7

YEAR_IN_NS = 3600 * 24 * 36525 * 10 ** 7


class DraftUUIDTests(unittest.TestCase):
    def test_uuid6_generation(self):
        uuid6_1 = uuid6()
        self.assertEqual(uuid6_1.version, 6)
        for _ in range(1000):
            uuid6_2 = uuid6()
            self.assertLess(uuid6_1, uuid6_2)
            uuid6_1 = uuid6_2

    def test_uuid7_generation(self):
        uuid7_1 = uuid7()
        self.assertEqual(uuid7_1.version, 7)
        for _ in range(1000):
            uuid7_2 = uuid7()
            self.assertLess(uuid7_1, uuid7_2)
            uuid7_1 = uuid7_2

    def test_invalid_int(self):
        with self.assertRaises(ValueError):
            _ = DraftUUID(int=-1)
        with self.assertRaises(ValueError):
            _ = DraftUUID(int=1 << 128)

    def test_valid_int(self):
        test_uuid = DraftUUID(int=0)
        self.assertEqual(test_uuid.version, None)
        self.assertEqual(test_uuid.time, 0)
        test_uuid = DraftUUID(int=(1 << 128) - 1)
        self.assertEqual(test_uuid.version, None)

    def test_invalid_version(self):
        with self.assertRaises(ValueError):
            _ = DraftUUID(int=1, version=420)

    @patch("uuid6._last_v7_timestamp", 1)
    @patch("time.time_ns", return_value=1234)
    def test_uuid7_same_nanosecond(self, mocktime):
        uuid7_1 = uuid7()
        for _ in range(10):
            uuid7_2 = uuid7()
            self.assertLess(uuid7_1, uuid7_2)
            self.assertEqual(uuid7_1.int >> 56, (uuid7_2.int >> 56) - 1)
            uuid7_1 = uuid7_2

    @patch("uuid6._last_v6_timestamp", 1)
    @patch("uuid6._getrandbits", return_value=678)
    @patch("time.time_ns", return_value=12345)
    def test_uuid6_fields_without_randomness(self, mocktime, mockrand):
        uuid6_1 = uuid6(clock_seq=123)
        for _ in range(10):
            uuid6_2 = uuid6(clock_seq=123)
            self.assertLess(uuid6_1, uuid6_2)
            self.assertEqual(uuid6_1.fields[0], uuid6_2.fields[0])
            self.assertEqual(uuid6_1.fields[1], uuid6_2.fields[1])
            self.assertEqual(uuid6_1.fields[2], uuid6_2.fields[2] - 1)
            self.assertEqual(uuid6_1.fields[3], uuid6_2.fields[3])
            self.assertEqual(uuid6_1.fields[4], uuid6_2.fields[4])
            self.assertEqual(uuid6_1.fields[5], uuid6_2.fields[5])
            uuid6_1 = uuid6_2

    @patch("uuid6._last_v6_timestamp", 1)
    def test_uuid6_far_in_future(self):
        with patch("time.time_ns", return_value=1):
            uuid_prev = uuid6()
        for i in range(1, 3260):
            with patch("time.time_ns", return_value=i * YEAR_IN_NS):
                uuid_cur = uuid6()
                self.assertLess(uuid_prev, uuid_cur)
                uuid_prev = uuid_cur

        # Overflow
        with patch("time.time_ns", return_value=3270 * YEAR_IN_NS):
            uuid_3270y_from_epoch = uuid6()
        self.assertLess(uuid_3270y_from_epoch, uuid_prev)

    @patch("uuid6._last_v7_timestamp", 1)
    def test_uuid7_far_in_future(self):
        with patch("time.time_ns", return_value=1):
            uuid_prev = uuid7()
        for i in range(1, 2170):
            with patch("time.time_ns", return_value=i * YEAR_IN_NS):
                uuid_cur = uuid7()
                self.assertLess(uuid_prev, uuid_cur)
                uuid_prev = uuid_cur

        # Overflow after 2 ** 36 seconds
        with patch("time.time_ns", return_value=2178 * YEAR_IN_NS):
            uuid_2178_from_epoch = uuid7()
        self.assertLess(uuid_2178_from_epoch, uuid_prev)

    def test_time(self):
        uuid_1 = uuid1()
        uuid_6 = uuid6()
        self.assertAlmostEqual(uuid_6.time / 10 ** 7, uuid_1.time / 10 ** 7, 3)
        cur_time = time_ns()
        uuid_7 = uuid7()
        self.assertAlmostEqual(uuid_7.time / 10 ** 9, cur_time / 10 ** 9, 3)


if __name__ == "__main__":
    unittest.main()



r"""UUID draft version objects (universally unique identifiers).
This module provides the functions uuid6() and uuid7() for
generating version 6 and 7 UUIDs as specified in
https://github.com/uuid6/uuid6-ietf-draft.
"""

from uuid import UUID


class DraftUUID(UUID):
    r"""UUID draft version objects"""

    def __init__(self, int: int, version: int = None) -> None:
        r"""Create a UUID from a single 128-bit integer as the 'int' argument."""

        if not 0 <= int < 1 << 128:
            raise ValueError("int is out of range (need a 128-bit value)")
        if version is not None:
            if not 6 <= version <= 7:
                raise ValueError("illegal version number")
            # Set the variant to RFC 4122.
            int &= ~(0xC000 << 48)
            int |= 0x8000 << 48
            # Set the version number.
            int &= ~(0xF000 << 64)
            int |= version << 76
        super().__init__(int=int)

    @property
    def subsec(self) -> int:
        return (
            (self.time_mid & 0x0FFF) << 18
            | (self.time_hi_version & 0x0FFF) << 6
            | self.clock_seq_hi_variant & 0x3F
        )

    @property
    def time(self) -> int:
        if self.version == 6:
            return (
                (self.time_low << 28)
                | (self.time_mid << 12)
                | (self.time_hi_version & 0x0FFF)
            )
        if self.version == 7:
            return self.unixts * 10 ** 9 + self.subsec
        return super().time

    @property
    def unixts(self) -> int:
        return self.time_low << 4 | self.time_mid >> 12


def _getrandbits(k: int) -> int:
    import secrets

    return secrets.randbits(k)


_last_v6_timestamp = None
_last_v7_timestamp = None


def uuid6(clock_seq: int = None) -> UUID:
    r"""Generate a UUID from sequence number, and the current time.
    If 'clock_seq' is given, it is used as the sequence number;
    otherwise a random 14-bit sequence number is chosen."""

    global _last_v6_timestamp
    import time

    nanoseconds = time.time_ns()
    # 0x01b21dd213814000 is the number of 100-ns intervals between the
    # UUID epoch 1582-10-15 00:00:00 and the Unix epoch 1970-01-01 00:00:00.
    timestamp = nanoseconds // 100 + 0x01B21DD213814000
    if _last_v6_timestamp is not None and timestamp <= _last_v6_timestamp:
        timestamp = _last_v6_timestamp + 1
    _last_v6_timestamp = timestamp
    if clock_seq is None:
        clock_seq = _getrandbits(14)  # instead of stable storage
    node = _getrandbits(48)
    time_high_and_time_mid = (timestamp >> 12) & 0xFFFFFFFFFFFF
    time_low_and_version = timestamp & 0x0FFF
    uuid_int = time_high_and_time_mid << 80
    uuid_int |= time_low_and_version << 64
    uuid_int |= (clock_seq & 0x3FFF) << 48
    uuid_int |= node
    return DraftUUID(int=uuid_int, version=6)


def uuid7() -> UUID:
    r"""The UUIDv7 format is designed to encode a Unix timestamp with
    arbitrary sub-second precision.  The key property provided by UUIDv7
    is that timestamp values generated by one system and parsed by
    another are guaranteed to have sub-second precision of either the
    generator or the parser, whichever is less.  Additionally, the system
    parsing the UUIDv7 value does not need to know which precision was
    used during encoding in order to function correctly."""

    global _last_v7_timestamp
    import time

    nanoseconds = time.time_ns()
    if _last_v7_timestamp is not None and nanoseconds <= _last_v7_timestamp:
        nanoseconds = _last_v7_timestamp + 1
    _last_v7_timestamp = nanoseconds
    timestamp_s, timestamp_ns = divmod(nanoseconds, 10 ** 9)
    subsec_a = timestamp_ns >> 18
    subsec_b = (timestamp_ns >> 6) & 0x0FFF
    subsec_c = timestamp_ns & 0x3F
    rand = _getrandbits(56)
    uuid_int = (timestamp_s & 0x0FFFFFFFFF) << 92
    uuid_int |= subsec_a << 80
    uuid_int |= subsec_b << 64
    uuid_int |= subsec_c << 56
    uuid_int |= rand
    return DraftUUID(int=uuid_int, version=7)



import os
from setuptools import setup, find_packages  # type: ignore

NAME = "uuid6"

DESCRIPTION = "New time-based UUID formats which are suited for use as a database key"
URL = "https://github.com/oittaa/uuid6-python"
LONG_DESCRIPTION = open(os.path.join(os.path.dirname(__file__), "README.md")).read()

AUTHOR = "Oittaa"
AUTHOR_EMAIL = ""
GITHUB_REF = os.environ.get("GITHUB_REF")
PREFIX = "refs/tags/"

if GITHUB_REF and GITHUB_REF.startswith(PREFIX):
    prefix_len = len(PREFIX)
    VERSION = GITHUB_REF[prefix_len:]
else:
    VERSION = "0.0.0.dev0"

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    url=URL,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    include_package_data=True,
    zip_safe=False,
    keywords=[
        "uuid",
        "uuid6",
        "uuid7",
        "uuidv6",
        "uuidv7",
    ],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
    ],
    setup_requires=[
        "wheel",
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3.7",
)
